@@ -95,15 +95,59 @@ GENERIC_RE = re.compile(r"""
 
 
 def resolve_generic_url(base_url, url_ref):
+    """Resolve a relative reference per RFC 3986
+
+    Used for the custom URI schemes supplied by
+    clients of the Mixer: HTTP URIs are processed
+    using the Python library’s urlparse module.
+
+    Arguments --
+        base_url -- absolute URL that is used
+            to interpret url_ref
+        url_ref -- may be an absolute URL
+            (which will be returned unchanged)
+            or a relative reference to be
+            resolved relative to base_url.
+
+    <http://tools.ietf.org/html/rfc3986>
+    """
+    if not url_ref:
+        return base_url
+
     if GENERIC_RE.match(url_ref):
         return url_ref
     m = GENERIC_RE.match(base_url)
     if url_ref.startswith('/'):
         return m.group('scheme') + '://' + m.group('authority') + url_ref
-    return (m.group('scheme')
-            + '://' + m.group('authority') + '/'
-            + (m.group('path') or '')
-            + url_ref)
+
+    scheme, authority, path = m.group('scheme'), m.group('authority'), (m.group('path') or '')
+
+    # Split reference up at slashes and strip out irrelevant .. and .:
+    elts = []
+    for elt in url_ref.split('/'):
+        if elt == '..' and elts and elts[-1] != '..':
+            elts.pop(-1)
+        else:
+            elts.append(elt)
+
+    # Now adjust path from base URL according to
+    # .. and . at the join.
+    while elts:
+        if elts[0] == '..':
+            if path:
+                path = path[:-1]
+                q = path.rfind('/')
+                if q >= 0:
+                    path = path[:q + 1]
+                else:
+                    path = ''
+            elts.pop(0)
+        elif elts[0] == '.':
+            elts.pop(0)
+        else:
+            break
+
+    return scheme + '://' + authority + '/' + path + '/'.join(elts)
 
 class CouldNotLoad(Exception): pass
 
@@ -126,6 +170,8 @@ class Loader(object):
 
         The spec can have a 'file' member specifying a file path.
         """
+        if isinstance(spec, basestring):
+            spec = {'href': spec}
         if 'file' in spec:
             file_path = resolve_file_path(spec['file'], base)
             return url_from_file_path(file_path)
@@ -186,7 +232,7 @@ class Loader(object):
         # If we get this far, we can’t work out a way to get a stream.
         return None
 
-    def get_url_stream(self, url):
+    def get_url_stream(self, url, ext='tpmaps'):
         """Given a URL, return input stream it specifies.
 
         Returns --
@@ -197,6 +243,10 @@ class Loader(object):
         """
         if url.startswith('file://'):
             file_path = url[7:]
+
+            if not os.path.exists(file_path) and os.path.exists(file_path + '.' + ext):
+                file_path += '.' + ext
+
             if os.path.isdir(file_path):
                 raise DudeItsADirectory(file_path)
 
@@ -235,7 +285,7 @@ class Loader(object):
         finally:
             strm.close()
 
-    def maybe_get_spec(self, spec, base=None):
+    def maybe_get_spec(self, spec, base=None, ext='tprx'):
         """Given a spec, return spec it refers to or the spec.
 
         The spec can have a 'file' member specifying a file path.
@@ -249,7 +299,7 @@ class Loader(object):
         spec = self._specs.get(url)
         if spec:
             return spec
-        meta, strm = self.get_url_stream(url)
+        meta, strm = self.get_url_stream(url, ext=ext)
         if meta['content-type'] == 'application/json':
             spec = json.load(strm)
         else:
@@ -263,17 +313,26 @@ class ResourceBase(object):
         self.name = name
 
     def get_bytes(self):
-        raise NotImplemented('{0}.get_bytes'.format(self.__class__.__name))
+        raise NotImplementedError('{0}.get_bytes'.format(self.__class__.__name__))
 
     def get_image(self):
-        raise NotImplemented('{0}.get_image'.format(self.__class__.__name))
+        raise NotImplementedError('{0}.get_image'.format(self.__class__.__name__))
+
+    def get_last_modified(self):
+        """When was the laast time this resource was changed?"""
+        raise NotImplementedError('{0}.get_last_modified'.format(self.__class__.__name__))
+
+    def is_modified_since(self, then):
+        """Has this resource been modified since this date?"""
+        return self.get_last_modified() > then
 
 
 class TextResource(ResourceBase):
     """A document whose content is a literal string."""
-    def __init__(self, name, content):
+    def __init__(self, name, content, last_modified=None):
         super(TextResource, self).__init__(name)
         self.content = content
+        self.last_modified = last_modified or datetime.now()
 
     def get_content(self):
         return self.content
@@ -282,6 +341,9 @@ class TextResource(ResourceBase):
         """Get the byte sequence that will go in the ZIP archive"""
         return self.get_content().encode('UTF-8')
 
+    def get_last_modified(self):
+        return self.last_modified
+
 
 class PackBase(object):
     """Base class for a texture pack.
@@ -304,22 +366,50 @@ class PackBase(object):
             ResoureBase subclass instance
 
         """
-        raise NotImplemented('{0}.get_resource'.format(self.__class__.__name__))
+        raise NotImplementedError('{0}.get_resource'.format(self.__class__.__name__))
 
     def get_resource_names(self):
         """A list of all resources in the pack."""
-        raise NotImplemented('{0}.get_resource_names'.format(self.__class__.__name__))
+        raise NotImplementedError('{0}.get_resource_names'.format(self.__class__.__name__))
 
     def write_to(self, strm):
         with ZipFile(strm, 'w', ZIP_DEFLATED) as zip:
             for name in sorted(self.get_resource_names()):
                 zip.writestr(name, self.get_resource(name).get_bytes())
 
+    def get_last_modified(self):
+        """Return a datetime object giving the last time a resource was modified.
+
+        Note that what matters is the modification times of
+        the resources in the pack. For example, two zips
+        created at different times but with the same timestamps
+        recorded for their contents will have the same last-modified time.
+
+        This is returned as an instance of datetime.
+        The precision will depend on the type of pack; for fiels
+        stored in a ZIP this is only to the nearest 2 seconds
+        (because of limitations in the ZIP format).
+        """
+        return max(self.get_resource(n).get_last_modified()
+                for n in self.get_resource_names())
+
+    def is_modified_since(self, then):
+        """Return true iff this pack has had a respource changed since the given date.
+
+        Because the precision of timestamps varies according
+        to the format the pack is stored in, near misses
+        will be unreliable except when the given date
+        is the same as the result of an earlier call to get_last_modified.
+        """
+        return any(self.get_resource(n).is_modified_since(then)
+                for n in self.get_resource_names())
+
+
 class RecipePack(PackBase):
     """A texture pack assembled from other resources."""
 
-    def __init__(self, label, desc):
-        super(RecipePack, self).__init__(Atlas())
+    def __init__(self, label, desc, atlas=None):
+        super(RecipePack, self).__init__(atlas or Atlas())
         self.label = label
         self.desc = desc
 
@@ -335,9 +425,15 @@ class RecipePack(PackBase):
     def get_resource_names(self):
         return self.resources.keys()
 
+    def is_modified_since(self, then):
+        for n in self.get_resource_names():
+            if n != 'pack.txt' and self.get_resource(n).is_modified_since(then):
+                return True
+        return False
+
 
 class SourcePack(PackBase):
-    """A texture pack that gets resources from a ZIP file."""
+    """A texture pack that gets resources from a ZIP file or directory."""
     def __init__(self, zip_data, atlas):
         super(SourcePack, self).__init__(atlas)
         if (isinstance(zip_data, basestring)
@@ -367,7 +463,8 @@ class SourcePack(PackBase):
             return res
         if name.endswith('.txt'):
             text = self.get_resource_bytes(name).decode('UTF-8')
-            res = TextResource(name, text)
+            last_modified = self.get_resource_last_modified(name)
+            res = TextResource(name, text, last_modified)
         else:
             res = SourceResource(self, name)
         self.loaded_resources[name] = res
@@ -433,6 +530,57 @@ class SourcePack(PackBase):
         return datetime(*ymdhms)
 
 
+class LazyPack(PackBase):
+    """A pack specified by a URL that will be loaded when required.
+
+    Actually performing the network operation will be deferred
+    until it can no longer be avoided."""
+    def __init__(self, loader, url, atlas):
+        super(LazyPack, self).__init__(atlas)
+        self.loader = loader
+        self.url = url
+        self.pack = None
+
+    def get_pack(self):
+        """If we have not done so already, download the pack."""
+        if not self.pack:
+            meta, strm = self.loader.get_url_stream(self.url, 'zip')
+            # XXX exception if wrong type of data?
+            self.pack = SourcePack(strm, self.atlas)
+        return self.pack
+
+    def get_resource(self, name):
+        """Get the named resource
+
+        Arguments --
+            name -- specifies the resource to return
+
+        Returns --
+            A resource (subclass of ResourceBase)
+        """
+        return self.get_pack().get_resource(name)
+
+    def get_resource_names(self):
+        return self.get_pack().get_resource_names()
+
+    @property
+    def label(self):
+        return self.get_pack().label
+
+    @property
+    def desc(self):
+        return self.get_pack().desc
+
+    def get_last_modified(self):
+        """Get a timestamp for the last time the content of the pack was changed.
+
+        Note that the contents timestamps are what matter;
+        the same files archived twice will have the same
+        last-modified time.
+        """
+        return self.get_pack().desc
+
+
 class SourceResource(ResourceBase):
     def __init__(self, source, name):
         self.source = source
@@ -466,6 +614,9 @@ class RenamedResource(ResourceBase):
     def get_image(self):
         return self.res.get_image()
 
+    def get_last_modified(self):
+        return self.res.get_last_modified()
+
 
 class MapBase(object):
         pass
@@ -595,6 +746,8 @@ class Atlas(object):
         # It had better be a list of maps if we get this far.
         return CompositeMap(self.get_map(x, base) for x in spec)
 
+    def get_map_names(self):
+        return self.maps.keys()
 
 def pil_box(left=None, top=None, right=None, bottom=None, x=None, y=None, width=None, height=None):
     """Return a bounding box of form (LEFT, TOP, RIGHT, BOTTOM) used by PIL.
@@ -618,7 +771,38 @@ def pil_box(left=None, top=None, right=None, bottom=None, x=None, y=None, width=
     return left, top, right, bottom
 
 
-class CompositeResource(ResourceBase):
+class ImagingResourceBase(ResourceBase):
+    """Base class for images that are lazily constructed using the Imaging library."""
+    def __init__(self, name):
+        super(ImagingResourceBase, self).__init__(name)
+        self._image = None
+        self._bytes = None
+
+    def _calc(self):
+        """Generate the image resource and return the image"""
+        raise NotImplementedError('{n}._calc'.format(n=self.__class__.__name__))
+
+    def _invalidate(self):
+        """Call this if the recipe changes, so cached images should be discarded."""
+        self._image = None
+        self._bytes = None
+
+    def get_image(self):
+        """Get the composite image."""
+        if self._image is None:
+            self._image = self._calc()
+        return self._image
+
+    def get_bytes(self):
+        """Get the bytes representing the composite image in PNG format."""
+        if self._bytes is None:
+            strm = StringIO()
+            self.get_image().save(strm, 'PNG', optimize=True)
+            self._bytes = strm.getvalue()
+        return self._bytes
+
+
+class CompositeResource(ImagingResourceBase):
     """An image made by replacing some cells in an image with parts of another"""
     def __init__(self, name, base_res, base_map):
         """Create a composite resouce (with no substitutions)
@@ -634,7 +818,6 @@ class CompositeResource(ResourceBase):
         self.res = base_res
         self.map = base_map
         self.replacements = []
-        self.image = None
 
     def replace(self, source_res, source_map, cells):
         """Replacing cell(s) in the image with cells from another
@@ -654,8 +837,7 @@ class CompositeResource(ResourceBase):
         self.replacements.append((source_res, source_map, cells))
 
         # This invalidates any cached image.
-        self.im = None
-        self.bytes = None
+        self._invalidate()
 
     def _calc(self):
         """Called when the image data is required.
@@ -669,21 +851,36 @@ class CompositeResource(ResourceBase):
                 src_box = src_map.get_box(src_name)
                 src_im = src_res.get_image().crop(src_box)
                 im.paste(src_im, dst_box)
-        self.image = im
+        return im
 
-    def get_image(self):
-        """Get the composite image."""
-        if self.image is None:
-            self._calc()
-        return self.image
+    def get_last_modified(self):
+        return max([self.res.get_last_modified()]
+            + [x.get_last_modified() for x, _, _ in self.replacements])
 
-    def get_bytes(self):
-        """Get the bytes representing the composite image in PNG format."""
-        if self.bytes is None:
-            strm = StringIO()
-            self.get_image().save(strm, 'PNG', optimize=True)
-            self.bytes = strm.getvalue()
-        return self.bytes
+
+class PackIconResource(ImagingResourceBase):
+    def __init__(self, res, map, names):
+        super(PackIconResource, self).__init__('pack.png')
+        self.res = res
+        self.map = map
+        self.names = names
+
+    def _calc(self):
+        src_im = self.res.get_image()
+        im = Image.new('RGBA', (128, 128), (255, 255, 255))
+        seq = [(0, 0, 64), (64, 0, 64), (64, 64, 64),
+            (32, 96, 32), (0, 96, 32), (0, 64, 32),
+            (32, 64, 16), (48, 64, 16), (48, 80, 16),
+            (32, 80, 16)]
+        for (x, y, wh), name in zip(seq, self.names):
+            dst_box = x, y, x + wh, y + wh
+            src_box = self.map.get_box(name)
+            cell_im = src_im.crop(src_box).resize((wh, wh))
+            im.paste(cell_im, dst_box)
+        return im
+
+    def get_last_modified(self):
+        return self.res.get_last_modified()
 
 
 class UnknownPack(Exception):
@@ -802,7 +999,10 @@ class Mixer(object):
                     yield res
             else:
                 # A recipe for creating a new resource from one or more sources.
-                res_name = file_spec['file']
+                if 'pack_icon' in file_spec:
+                    res_name = file_spec.get('file', 'pack.png')
+                else:
+                    res_name = file_spec['file']
                 src_res = src_pack.get_resource(file_spec.get('source', res_name))
                 if 'replace' in file_spec:
                     src_map = self.get_map(src_pack.atlas, file_spec.get('map', src_res.name), base)
@@ -820,6 +1020,9 @@ class Mixer(object):
                         res.replace(src2_res, src2_map, cells)
                 elif res_name == src_res.name:
                     res = src_res
+                elif 'pack_icon' in file_spec:
+                    src_map = self.get_map(src_pack.atlas, file_spec.get('map', src_res.name), base)
+                    res = PackIconResource(src_res, src_map, file_spec['pack_icon']['cells'])
                 else:
                     res = RenamedResource(res_name, src_res)
                 yield res
@@ -845,6 +1048,10 @@ class Mixer(object):
                     data -- value must be a ZIP archive;
                     base64 -- value must be a ZIP archive,
                         converted to ASCII with the base64 codec.
+                    unjumble -- assume the files in the pack
+                        have the right names byt the wrong folders
+                        and use the atlass to determine which files
+                        to pull out of the ZIP
             fallback_pack (optional) --
                 If the pack_spec is the empty string or None,
                 then use this value instead.
@@ -868,11 +1075,29 @@ class Mixer(object):
                 raise UnknownPack(pack_spec)
             return result
 
-        atlas = self.get_atlas(hasattr(pack_spec, 'get') and pack_spec.get('maps'), base)
+        atlas_spec = hasattr(pack_spec, 'get') and pack_spec.get('maps', pack_spec.get('unjumble'))
+        atlas = self.get_atlas(atlas_spec, base)
+
+        pack = self._load_raw_pack(atlas, pack_spec, base)
+        if hasattr(pack_spec, 'get'):
+            unjumble_spec = pack_spec.get('unjumble')
+            if unjumble_spec:
+                unjumble_atlas = atlas if 'maps' not in pack_spec else self.get_atlas(unjumble_spec, base)
+                pack = self.make_unjumbled_pack(pack, unjumble_atlas)
+        return pack
+
+    def _load_raw_pack(self, atlas, pack_spec, base):
+
+        # Various ways of decoding the pack spec.
+        url = self.loader.get_url(pack_spec, base)
+        if url and url.startswith('http://'):
+            return LazyPack(self.loader, url, atlas)
 
-        # Varfious ways of decoding the pack spec.
         try:
-            strm = self.loader.get_stream(pack_spec, base)
+            if url:
+                meta, strm = self.loader.get_url_stream(url)
+            else:
+                strm = self.loader.get_stream(pack_spec, base)
         except CouldNotLoad, e:
             raise UnknownPack(pack_spec, e)
         except DudeItsADirectory, e:
@@ -884,6 +1109,27 @@ class Mixer(object):
 
         return SourcePack(strm, atlas)
 
+    def make_unjumbled_pack(self, pack, atlas):
+        """Make an unjumbled copy of this pack using this atlas.
+
+        Given a pack with correct file names but possibly wrong folder names
+        create a new pack that rearranges the files to conform
+        to texturepack conventions.
+        """
+        pack_names = list(pack.get_resource_names())
+        new_pack = RecipePack('unjumbled', 'unjumbled', atlas=atlas)
+        for desired_name in atlas.get_map_names():
+            nick = desired_name.split('/')[-1]
+            slash_nick = '/' + nick
+            for i, available_name in enumerate(pack_names):
+                if available_name == nick or available_name.endswith(slash_nick):
+                    res = pack.get_resource(available_name)
+                    if desired_name != available_name:
+                        res = RenamedResource(desired_name, res)
+                    new_pack.add_resource(res)
+                    del pack_names[i]
+                    break
+        return new_pack
 
     def get_map(self, atlas, spec, base):
         """Get a map from the pack if possible, otherwise try the global atlas."""
@@ -917,10 +1163,10 @@ class Mixer(object):
                 if 'file' in atlas_spec
                 else self.loader.get_url(atlas_spec, base))
             if hasattr(atlas_spec, 'items'):
-                atlas_spec = self.loader.maybe_get_spec(atlas_spec, base)
+                atlas_spec = self.loader.maybe_get_spec(atlas_spec, base, 'tpmaps')
             if hasattr(atlas_spec, 'items'):
                 for name, map_spec in atlas_spec.items():
-                    atlas.add_map(name, self.get_map(atlas, map_spec, atlas_base))
+                    atlas.add_map(name, map_spec and self.get_map(atlas, map_spec, atlas_base))
             else:
                 # Atlas is a list of atlasses to be merged.
                 for spec in atlas_spec:
